Currently to fill a struct pathspec, we do:

   const char **paths;
   paths = get_pathspec(prefix, argv);
   ...
   init_pathspec(&pathspec, paths);

"paths" can only carry bare strings, which loses information from
command line arguments such as pathspec magic or the prefix part's
length for each argument.

parse_pathspec() is introduced to combine the two calls into one. The
plan is gradually replace all get_pathspec() and init_pathspec() with
parse_pathspec(). get_pathspec() now becomes a thin wrapper of
parse_pathspec().

parse_pathspec() allows the caller to reject the pathspec magics that
it does not support. When a new pathspec magic is introduced, we can
enable it per command after making sure that all underlying code has no
problem with the new magic.

"flags" parameter is currently unused. But it would allow callers to
pass certain instructions to parse_pathspec, for example forcing
literal pathspec when no magic is used.

With the introduction of parse_pathspec, there are now two functions
that can initialize struct pathspec: init_pathspec and
parse_pathspec. Any semantic changes in struct pathspec must be
reflected in both functions. init_pathspec() will be phased out in
favor of parse_pathspec().

Signed-off-by: Nguyễn Thái Ngọc Duy <pclo...@gmail.com>
---
 dir.c   |  2 +-
 dir.h   |  1 +
 setup.c | 99 +++++++++++++++++++++++++++++++++++++++++++++++++----------------
 3 files changed, 77 insertions(+), 25 deletions(-)

diff --git a/dir.c b/dir.c
index c391d46..31f0995 100644
--- a/dir.c
+++ b/dir.c
@@ -291,7 +291,7 @@ int match_pathspec_depth(const struct pathspec *ps,
 /*
  * Return the length of the "simple" part of a path match limiter.
  */
-static int simple_length(const char *match)
+int simple_length(const char *match)
 {
        int len = -1;
 
diff --git a/dir.h b/dir.h
index f5c89e3..1d4888b 100644
--- a/dir.h
+++ b/dir.h
@@ -66,6 +66,7 @@ struct dir_struct {
 #define MATCHED_RECURSIVELY 1
 #define MATCHED_FNMATCH 2
 #define MATCHED_EXACTLY 3
+extern int simple_length(const char *match);
 extern char *common_prefix(const char **pathspec);
 extern int match_pathspec(const char **pathspec, const char *name, int 
namelen, int prefix, char *seen);
 extern int match_pathspec_depth(const struct pathspec *pathspec,
diff --git a/setup.c b/setup.c
index f108c4b..4fcdae6 100644
--- a/setup.c
+++ b/setup.c
@@ -174,7 +174,7 @@ static struct pathspec_magic {
 
 /*
  * Take an element of a pathspec and check for magic signatures.
  * Append the result to the prefix.
  *
  * For now, we only parse the syntax and throw out anything other than
  * "top" magic.
@@ -185,7 +185,10 @@ static struct pathspec_magic {
  * the prefix part must always match literally, and a single stupid
  * string cannot express such a case.
  */
-static const char *prefix_pathspec(const char *prefix, int prefixlen, const 
char *elt)
+static unsigned prefix_pathspec(struct pathspec_item *item,
+                               const char **raw,
+                               const char *prefix, int prefixlen,
+                               const char *elt)
 {
        unsigned magic = 0;
        const char *copyfrom = elt;
@@ -241,39 +244,87 @@ static const char *prefix_pathspec(const char *prefix, 
int prefixlen, const char
        }
 
        if (magic & PATHSPEC_FROMTOP)
-               return xstrdup(copyfrom);
+               item->match = xstrdup(copyfrom);
        else
-               return prefix_path(prefix, prefixlen, copyfrom);
+               item->match = prefix_path(prefix, prefixlen, copyfrom);
+       *raw = item->match;
+       item->len = strlen(item->match);
+       item->nowildcard_len = simple_length(item->match);
+       return magic;
 }
 
-const char **get_pathspec(const char *prefix, const char **pathspec)
+static int pathspec_item_cmp(const void *a_, const void *b_)
 {
-       const char *entry = *pathspec;
-       const char **src, **dst;
-       int prefixlen;
+       struct pathspec_item *a, *b;
 
-       if (!prefix && !entry)
-               return NULL;
+       a = (struct pathspec_item *)a_;
+       b = (struct pathspec_item *)b_;
+       return strcmp(a->match, b->match);
+}
+
+/*
+ * Given command line arguments and a prefix, convert the input to
+ * pathspec. die() if any magic other than ones in magic_mask.
+ */
+static void parse_pathspec(struct pathspec *pathspec,
+                          unsigned magic_mask, unsigned flags,
+                          const char *prefix, const char **argv)
+{
+       struct pathspec_item *item;
+       const char *entry = *argv;
+       int i, n, prefixlen;
+
+       memset(pathspec, 0, sizeof(*pathspec));
+
+       /* No arguments, no prefix -> no pathspec */
+       if (!entry && !prefix)
+               return;
 
+       /* No arguments with prefix -> prefix pathspec */
        if (!entry) {
-               static const char *spec[2];
-               spec[0] = prefix;
-               spec[1] = NULL;
-               return spec;
+               static const char *raw[2];
+
+               pathspec->items = item = xmalloc(sizeof(*item));
+               item->match = prefix;
+               item->nowildcard_len = item->len = strlen(prefix);
+               raw[0] = prefix;
+               raw[1] = NULL;
+               pathspec->nr = 1;
+               pathspec->raw = raw;
+               return;
        }
 
-       /* Otherwise we have to re-write the entries.. */
-       src = pathspec;
-       dst = pathspec;
+       n = 0;
+       while (argv[n])
+               n++;
+
+       pathspec->nr = n;
+       pathspec->items = item = xmalloc(sizeof(*item) * n);
+       pathspec->raw = argv;
        prefixlen = prefix ? strlen(prefix) : 0;
-       while (*src) {
-               *(dst++) = prefix_pathspec(prefix, prefixlen, *src);
-               src++;
+
+       for (i = 0; i < n; i++) {
+               unsigned applied_magic;
+               const char *arg = argv[i];
+
+               applied_magic = prefix_pathspec(item + i, argv + i,
+                                               prefix, prefixlen, arg);
+               if (applied_magic & ~magic_mask)
+                       die(_("pathspec magic in '%s' is not supported"
+                             " by this command"), arg);
+               if (item[i].nowildcard_len < item[i].len)
+                       pathspec->has_wildcard = 1;
        }
-       *dst = NULL;
-       if (!*pathspec)
-               return NULL;
-       return pathspec;
+
+       qsort(pathspec->items, pathspec->nr,
+             sizeof(struct pathspec_item), pathspec_item_cmp);
+}
+
+const char **get_pathspec(const char *prefix, const char **pathspec)
+{
+       struct pathspec ps;
+       parse_pathspec(&ps, PATHSPEC_FROMTOP, 0, prefix, pathspec);
+       return ps.raw;
 }
 
 /*
@@ -174,7 +174,7 @@ static struct pathspec_magic {
 /*
  * Take an element of a pathspec and check for magic signatures.
- * Append the result to the prefix.
+ * Append the result to the prefix. Return the magic bitmap.
  *
  * For now, we only parse the syntax and throw out anything other than
-- 
1.8.0.rc2.23.g1fb49df

--
To unsubscribe from this list: send the line "unsubscribe git" in
the body of a message to majord...@vger.kernel.org
More majordomo info at  http://vger.kernel.org/majordomo-info.html

Reply via email to